//! HdrSample is a port of Gil Tene's HdrHistogram to native Rust. It provides recording and
//! analyzing of sampled data value counts across a large, configurable value range with
//! configurable precision within the range. The resulting "HDR" histogram allows for fast and
//! accurate analysis of the extreme ranges of data with non-normal distributions, like latency.
//!
//! # HdrHistogram
//!
//! What follows is a description from [the HdrHistogram
//! website](https://hdrhistogram.github.io/HdrHistogram/). Users are encourages to read the
//! documentation from the original [Java
//! implementation](https://github.com/HdrHistogram/HdrHistogram), as most of the concepts
//! translate directly to the Rust port.
//!
//! HdrHistogram supports the recording and analyzing of sampled data value counts across a
//! configurable integer value range with configurable value precision within the range. Value
//! precision is expressed as the number of significant digits in the value recording, and provides
//! control over value quantization behavior across the value range and the subsequent value
//! resolution at any given level.
//!
//! For example, a Histogram could be configured to track the counts of observed integer values
//! between 0 and 3,600,000,000 while maintaining a value precision of 3 significant digits across
//! that range. Value quantization within the range will thus be no larger than 1/1,000th (or 0.1%)
//! of any value. This example Histogram could be used to track and analyze the counts of observed
//! response times ranging between 1 microsecond and 1 hour in magnitude, while maintaining a value
//! resolution of 1 microsecond up to 1 millisecond, a resolution of 1 millisecond (or better) up
//! to one second, and a resolution of 1 second (or better) up to 1,000 seconds. At it's maximum
//! tracked value (1 hour), it would still maintain a resolution of 3.6 seconds (or better).
//!
//! HDR Histogram is designed for recoding histograms of value measurements in latency and
//! performance sensitive applications. Measurements show value recording times as low as 3-6
//! nanoseconds on modern (circa 2014) Intel CPUs. The HDR Histogram maintains a fixed cost in both
//! space and time. A Histogram's memory footprint is constant, with no allocation operations
//! involved in recording data values or in iterating through them. The memory footprint is fixed
//! regardless of the number of data value samples recorded, and depends solely on the dynamic
//! range and precision chosen. The amount of work involved in recording a sample is constant, and
//! directly computes storage index locations such that no iteration or searching is ever involved
//! in recording data values.
//!
//! # Interacting with the library
//!
//! HdrSample's API follows that of the original HdrHistogram Java implementation, with some
//! modifications to make its use more idiomatic in Rust. The description in this section has been
//! adapted from that given by the [Python port](https://github.com/HdrHistogram/HdrHistogram_py),
//! as it gives a nicer first-time introduction to the use of HdrHistogram than the Java docs do.
//!
//! HdrSample is generally used in one of two modes: recording samples, or querying for analytics.
//! In distributed deployments, the recording may be performed remotely (and possibly in multiple
//! locations), to then be aggregated later in a central location for analysis.
//!
//! ## Recording samples
//!
//! A histogram instance is created using the `::new` methods on the `Histogram` struct. These come
//! in three variants: `new`, `new_with_max`, and `new_with_bounds`. The first of these only sets
//! the required precision of the sampled data, but leaves the value range open such that any value
//! may be recorded. A `Histogram` created this way (or one where auto-resize has been explicitly
//! enabled) will automatically resize itself if a value that is too large to fit in the current
//! dataset is encountered. `new_with_max` sets an upper bound on the values to be recorded, and
//! disables auto-resizing, thus preventing any re-allocation during recording. If the application
//! attempts to record a larger value than this maximum bound, the record call will fail. Finally,
//! `new_with_bounds` restricts the lowest representible value of the dataset, such that a smaller
//! range needs to be covered (thus reducing the overall allocation size).
//!
//! For example the example below shows how to create a `Histogram` that can count values in the
//! `[1..3600000]` range with 1% precision, which could be used to track latencies in the range `[1
//! msec..1 hour]`).
//!
//! ```
//! use hdrsample::Histogram;
//! let mut hist = Histogram::<u64>::new_with_bounds(1, 60 * 60 * 1000, 2).unwrap();
//!
//! // samples can be recorded using .record, which will error if the value is too small or large
//! hist.record(54321).expect("value 54321 should be in range");
//!
//! // for ergonomics, samples can also be recorded with +=
//! // this call will panic if the value is out of range!
//! hist += 54321;
//!
//! // if the code that generates the values is subject to Coordinated Omission,
//! // the self-correcting record method should be used instead.
//! // for example, if the expected sampling interval is 10 msec:
//! hist.record_correct(54321, 10).expect("value 54321 should be in range");
//! ```
//!
//! Note the `u64` annotation. This type can be changed to reduce the storage overhead for all the
//! histogram bins, at the cost of a risk of overflowing if a large number of samples end up in the
//! same bin.
//!
//! ## Querying samples
//!
//! At any time, the histogram can be queried to return interesting statistical measurements, such
//! as the total number of recorded samples, or the value at a given percentile:
//!
//! ```
//! use hdrsample::Histogram;
//! let hist = Histogram::<u64>::new(2).unwrap();
//! // ...
//! println!("# of samples: {}", hist.count());
//! println!("99.9'th percentile: {}", hist.value_at_percentile(99.9));
//! ```
//!
//! Several useful iterators are also provided for quickly getting an overview of the dataset. The
//! simplest one is `iter_recorded()`, which yields one item for every non-empty sample bin. All
//! the HdrHistogram iterators are supported in HdrSample, so look for the `*Iterator` classes in
//! the [Java documentation](https://hdrhistogram.github.io/HdrHistogram/JavaDoc/).
//!
//! ```
//! use hdrsample::Histogram;
//! let hist = Histogram::<u64>::new(2).unwrap();
//! // ...
//! for (value, percentile, _, count) in hist.iter_recorded() {
//!     println!("{}'th percentile of data is {} with {} samples", percentile, value, count);
//! }
//! ```
//!
//! # Limitations and Caveats
//!
//! As with all the other HdrHistogram ports, the latest features and bug fixes from the upstream
//! HdrHistogram implementations may not be available in this port. A number of features have also
//! not (yet) been implemented:
//!
//!  - Concurrency support (`AtomicHistogram`, `ConcurrentHistogram`, …).
//!  - `DoubleHistogram`. You can use `f64` as the counter type, but none of the "special"
//!    `DoubleHistogram` features are supported.
//!  - The `Recorder` feature of HdrHistogram.
//!  - Value shifting ("normalization").
//!  - Histogram serialization and encoding/decoding.
//!  - Timestamps and tags.
//!  - Textual output methods. These seem almost orthogonal to HdrSample, though it might be
//!    convenient if we implemented some relevant traits (CSV, JSON, and possibly simple
//!    `fmt::Display`).
//!
//! Most of these should be fairly straightforward to add, as the code aligns pretty well with the
//! original Java/C# code. If you do decide to implement one and send a PR, please make sure you
//! also port the [test
//! cases](https://github.com/HdrHistogram/HdrHistogram/tree/master/src/test/java/org/HdrHistogram),
//! and try to make sure you implement appropriate traits to make the use of the feature as
//! ergonomic as possible.

#![allow(non_snake_case)]
#![deny(missing_docs)]

extern crate num;

use std::ops::Index;
use std::ops::IndexMut;
use std::ops::AddAssign;
use std::ops::SubAssign;
use std::borrow::Borrow;

/// `Histogram` is the core data structure in HdrSample. It records values, and performs analytics.
///
/// At its heart, it keeps the count for recorded samples in "buckets" of values. The resolution
/// and distribution of these buckets is tuned based on the desired highest trackable value, as
/// well as the user-specified number of significant decimal digits to preserve. The values for the
/// buckets are kept in a way that resembles floats and doubles: there is a mantissa and an
/// exponent, and each bucket represents a different exponent. The "sub-buckets" within a bucket
/// represent different values for the mantissa.
///
/// To a first approximation, the sub-buckets of the first
/// bucket would hold the values `0`, `1`, `2`, `3`, …, the sub-buckets of the second bucket would
/// hold `0`, `2`, `4`, `6`, …, the third would hold `0`, `4`, `8`, and so on. However, the low
/// half of each bucket (except bucket 0) is unnecessary, since those values are already covered by
/// the sub-buckets of all the preceeding buckets. Thus, `Histogram` keeps the top half of every
/// such bucket.
///
/// For the purposes of explanation, consider a `Histogram` with 2048 sub-buckets for every bucket,
/// and a lowest discernible value of 1:
///
/// <pre>
/// The 0th bucket covers 0...2047 in multiples of 1, using all 2048 sub-buckets
/// The 1st bucket covers 2048..4097 in multiples of 2, using only the top 1024 sub-buckets
/// The 2nd bucket covers 4096..8191 in multiple of 4, using only the top 1024 sub-buckets
/// ...
/// </pre>
///
/// Bucket 0 is "special" here. It is the only one that has 2048 entries. All the rest have
/// 1024 entries (because their bottom half overlaps with and is already covered by the all of
/// the previous buckets put together). In other words, the `k`'th bucket could represent `0 *
/// 2^k` to `2048 * 2^k` in 2048 buckets with `2^k` precision, but the midpoint of `1024 * 2^k
/// = 2048 * 2^(k-1)`, which is the k-1'th bucket's end. So, we would use the previous bucket
/// for those lower values as it has better precision.
///
pub struct Histogram<T: num::Num + num::ToPrimitive + Copy> {
    autoResize: bool,

    highestTrackableValue: i64,
    lowestDiscernibleValue: i64,
    significantValueDigits: u32,

    bucketCount: usize,
    subBucketCount: usize,

    leadingZeroCountBase: u32,
    subBucketHalfCountMagnitude: usize,

    unitMagnitude: usize,
    subBucketHalfCount: usize,

    subBucketMask: i64,

    unitMagnitudeMask: i64,

    // TODO: use options for these?
    maxValue: i64, // 0
    minNonZeroValue: i64, // MAX

    // atomics for maxValue/minNonZeroValue ?
    totalCount: i64,
    counts: Vec<T>,
}

/// Module containing the implementations of all `Histogram` iterators.
pub mod iterators;

impl<T: num::Num + num::ToPrimitive + Copy> Histogram<T> {
    // ********************************************************************************************
    // Histogram administrative read-outs
    // ********************************************************************************************

    /// Get the current number of distinct counted values in the histogram.
    pub fn len(&self) -> usize {
        self.counts.len()
    }

    /// Get the lowest discernible value for the histogram in its current configuration.
    pub fn low(&self) -> i64 {
        self.lowestDiscernibleValue
    }

    /// Get the highest trackable value for the histogram in its current configuration.
    pub fn high(&self) -> i64 {
        self.highestTrackableValue
    }

    /// Get the number of significant value digits kept by this histogram.
    pub fn sigfig(&self) -> u32 {
        self.significantValueDigits
    }

    /// Get the total number of samples recorded.
    pub fn count(&self) -> i64 {
        self.totalCount
    }

    /// Get the index of the last histogram bin.
    pub fn last(&self) -> usize {
        self.len() - 1
    }

    /// Get the number of buckets used by the histogram to cover the highest trackable value.
    ///
    /// This method differs from `.len()` in that it does not count the sub buckets within each
    /// bucket.
    ///
    /// This method is probably only useful for testing purposes.
    pub fn buckets(&self) -> usize {
        self.bucketCount
    }

    // ********************************************************************************************
    // Methods for looking up the count for a given value/index
    // ********************************************************************************************

    /// Find the bucket the given value should be placed in.
    ///
    /// May panic if the given value falls outside the current range of the histogram.
    fn index_for(&self, value: i64) -> isize {
        if value < 0 {
            panic!("Histogram recorded value cannot be negative.");
        }

        let bucketIndex = self.bucket_for(value);
        let subBucketIndex = self.sub_bucket_for(value, bucketIndex);

        debug_assert!(subBucketIndex < self.subBucketCount);
        debug_assert!(bucketIndex == 0 || (subBucketIndex >= self.subBucketHalfCount));

        // Calculate the index for the first entry that will be used in the bucket (halfway through
        // subBucketCount). For bucketIndex 0, all subBucketCount entries may be used, but
        // bucketBaseIndex is still set in the middle.
        let bucketBaseIndex = (bucketIndex + 1) << self.subBucketHalfCountMagnitude;

        // Calculate the offset in the bucket. This subtraction will result in a positive value in
        // all buckets except the 0th bucket (since a value in that bucket may be less than half
        // the bucket's 0 to subBucketCount range). However, this works out since we give bucket 0
        // twice as much space.
        let offsetInBucket = subBucketIndex as isize - self.subBucketHalfCount as isize;

        // The following is the equivalent of
        // ((subBucketIndex  - subBucketHalfCount) + bucketBaseIndex;
        (bucketBaseIndex as isize + offsetInBucket)
    }

    /// Get a mutable reference to the count bucket for the given value, if it is in range.
    fn mut_at(&mut self, value: i64) -> Option<&mut T> {
        let i = self.index_for(value);
        if i < 0 {
            return None;
        }

        self.counts.get_mut(i as usize)
    }

    // ********************************************************************************************
    // Histograms should be cloneable.
    // ********************************************************************************************

    /// Get a copy of this histogram, corrected for coordinated omission.
    ///
    /// To compensate for the loss of sampled values when a recorded value is larger than the
    /// expected interval between value samples, the new histogram will include an auto-generated
    /// additional series of decreasingly-smaller (down to the `interval`) value records for each
    /// count found in the current histogram that is larger than the `interval`.
    ///
    /// Note: This is a post-correction method, as opposed to the at-recording correction method
    /// provided by `record_correct`. The two methods are mutually exclusive, and only one of the
    /// two should be be used on a given data set to correct for the same coordinated omission
    /// issue.
    ///
    /// See notes in the description of the Histogram calls for an illustration of why this
    /// corrective behavior is important.
    ///
    /// If `interval` is larger than 0, add auto-generated value records as appropriate if value is
    /// larger than `interval`.
    pub fn clone_correct(&self, interval: i64) -> Histogram<T> {
        let mut h = Histogram::new_from(self);
        for (value, _, count, _) in self.iter_recorded() {
            h.record_n_correct(value, count, interval).unwrap();
        }
        h
    }

    /// Overwrite this histogram with the given histogram. All data and statistics in this
    /// histogram will be overwritten.
    pub fn set_to<B: Borrow<Histogram<T>>>(&mut self, source: B) -> Result<(), &'static str> {
        self.reset();
        self.add(source.borrow())
    }

    /// Overwrite this histogram with the given histogram while correcting for coordinated
    /// omission. All data and statistics in this histogram will be overwritten. See
    /// `clone_correct` for more detailed explanation about how correction is applied
    pub fn set_to_corrected<B: Borrow<Histogram<T>>>(&mut self,
                                                     source: B,
                                                     interval: i64)
                                                     -> Result<(), ()> {
        self.reset();
        self.add_correct(source, interval)
    }

    // ********************************************************************************************
    // Add and subtract methods for, well, adding or subtracting two histograms
    // ********************************************************************************************

    /// Add the contents of another histogram to this one.
    ///
    /// May fail if values in the other histogram are higher than `.high()`, and auto-resize is
    /// disabled.
    pub fn add<B: Borrow<Histogram<T>>>(&mut self, source: B) -> Result<(), &'static str> {
        let source = source.borrow();

        // make sure we can take the values in source
        let top = self.highest_equivalent(self.value_for(self.last()));
        if top < source.max() {
            if !self.autoResize {
                return Err("The other histogram includes values that do not fit in this \
                            histogram's range.");
            }
            self.resize(source.max());
        }

        if self.bucketCount == source.bucketCount && self.subBucketCount == source.subBucketCount &&
           self.unitMagnitude == source.unitMagnitude {
            // Counts arrays are of the same length and meaning,
            // so we can just iterate and add directly:
            let mut observedOtherTotalCount = 0i64;
            for i in 0..source.len() {
                let otherCount = source[i];
                if otherCount != T::zero() {
                    self[i] = self[i] + otherCount;
                    observedOtherTotalCount += otherCount.to_i64().unwrap();
                }
            }

            self.totalCount += observedOtherTotalCount;
            let mx = source.max();
            if mx > self.max() {
                self.update_max(mx);
            }
            let mn = source.min_nz();
            if mn < self.min_nz() {
                self.update_min(mn);
            }
        } else {
            // Arrays are not a direct match (or the other could change on the fly in some valid
            // way), so we can't just stream through and add them. Instead, go through the array
            // and add each non-zero value found at it's proper value:

            // Do max value first, to avoid max value updates on each iteration:
            let otherMaxIndex = source.index_for(source.max()) as usize;
            let otherCount = source[otherMaxIndex];
            self.record_n(source.value_for(otherMaxIndex), otherCount).unwrap();

            // Record the remaining values, up to but not including the max value:
            for i in 0..otherMaxIndex {
                let otherCount = source[i];
                if otherCount != T::zero() {
                    self.record_n(source.value_for(i), otherCount).unwrap();
                }
            }
        }

        // TODO:
        // if source.startTime < self.startTime {
        //     self.startTime = source.startTime;
        // }
        // if source.endTime > self.endTime {
        //     self.endTime = source.endTime;
        // }
        Ok(())
    }

    /// Add the contents of another histogram to this one, while correcting for coordinated
    /// omission.
    ///
    /// To compensate for the loss of sampled values when a recorded value is larger than the
    /// expected interval between value samples, the values added will include an auto-generated
    /// additional series of decreasingly-smaller (down to the given `interval`) value records for
    /// each count found in the current histogram that is larger than `interval`.
    ///
    /// Note: This is a post-recording correction method, as opposed to the at-recording correction
    /// method provided by `record_correct`. The two methods are mutually exclusive, and only one
    /// of the two should be be used on a given data set to correct for the same coordinated
    /// omission issue.
    ///
    /// See notes in the description of the `Histogram` calls for an illustration of why this
    /// corrective behavior is important.
    ///
    /// May fail if values in the other histogram are higher than `.high()`, and auto-resize is
    /// disabled.
    pub fn add_correct<B: Borrow<Histogram<T>>>(&mut self,
                                                source: B,
                                                interval: i64)
                                                -> Result<(), ()> {
        let source = source.borrow();

        for (value, _, count, _) in source.iter_recorded() {
            try!(self.record_n_correct(value, count, interval));
        }
        Ok(())
    }

    /// Subtract the contents of another histogram from this one.
    ///
    /// May fail if values in the other histogram are higher than `.high()`, and auto-resize is
    /// disabled. Or, if the count for a given value in the other histogram is higher than that of
    /// this histogram. In the latter case, some of the counts may still have been updated, which
    /// may cause data corruption.
    pub fn subtract<B: Borrow<Histogram<T>>>(&mut self, other: B) -> Result<(), &'static str> {
        let other = other.borrow();

        // make sure we can take the values in source
        let top = self.highest_equivalent(self.value_for(self.last()));
        if top < other.max() {
            if !self.autoResize {
                return Err("The other histogram includes values that do not fit in this \
                            histogram's range.");
            }
            self.resize(other.max());
        }

        for i in 0..other.len() {
            let otherCount = other[i];
            if otherCount != T::zero() {
                let otherValue = other.value_for(i);
                if self.count_at(otherValue).unwrap().to_i64().unwrap() <
                   otherCount.to_i64().unwrap() {
                    return Err("The other histogram includes counts that are higher than the \
                                current count for that value.");
                }
                self.alter_n(otherValue, otherCount, false).expect("value should fit by now");
            }
        }

        // With subtraction, the max and minNonZero values could have changed:
        if self.count_at(self.max()).unwrap() == T::zero() ||
           self.count_at(self.min_nz()).unwrap() == T::zero() {
            let l = self.len();
            self.restat(l);
        }

        Ok(())
    }

    // ********************************************************************************************
    // Setters and resetters.
    // ********************************************************************************************

    /// Clear the contents of this histogram while preserving its statistics and configuration.
    pub fn clear(&mut self) {
        for c in self.counts.iter_mut() {
            *c = T::zero();
        }
        self.totalCount = 0;
    }

    /// Reset the contents and statistics of this histogram, preserving only its configuration.
    pub fn reset(&mut self) {
        self.clear();

        self.reset_max(0);
        self.reset_min(i64::max_value());
        // self.normalizingIndexOffset = 0;
        // self.startTime = time::Instant::now();
        // self.endTime = time::Instant::now();
        // self.tag = String::new();
    }

    /// Control whether or not the histogram can auto-resize and auto-adjust it's highest trackable
    /// value as high-valued samples are recorded.
    pub fn auto(&mut self, enabled: bool) {
        self.autoResize = enabled;
    }

    // ********************************************************************************************
    // Construction.
    // ********************************************************************************************

    /// Construct an auto-resizing `Histogram` with a lowest discernible value of 1 and an
    /// auto-adjusting highest trackable value. Can auto-resize up to track values up to
    /// `(i64::max_value() / 2)`.
    ///
    /// `sigfig` specifies the number of significant value digits to preserve in the recorded data.
    /// This is the number of significant decimal digits to which the histogram will maintain value
    /// resolution and separation. Must be a non-negative integer between 0 and 5.
    pub fn new(sigfig: u32) -> Result<Histogram<T>, &'static str> {
        let mut h = Self::new_with_bounds(1, 2, sigfig);
        if let Ok(ref mut h) = h {
            h.autoResize = true;
        }
        h
    }

    /// Construct a `Histogram` given a known maximum value to be tracked, and a number of
    /// significant decimal digits. The histogram will be constructed to implicitly track
    /// (distinguish from 0) values as low as 1. Auto-resizing will be disabled.
    ///
    /// `high` is the highest value to be tracked by the histogram, and must be a positive integer
    /// that is >= 2. `sigfig` specifies the number of significant figures to maintain. This is the
    /// number of significant decimal digits to which the histogram will maintain value resolution
    /// and separation. Must be a non-negative integer between 0 and 5.
    pub fn new_with_max(high: i64, sigfig: u32) -> Result<Histogram<T>, &'static str> {
        Self::new_with_bounds(1, high, sigfig)
    }

    /// Construct a `Histogram` with known upper and lower bounds for recorded sample values.
    /// Providing a lowest discernible value (`low`) is useful is situations where the units used
    /// for the histogram's values are much smaller that the minimal accuracy required. E.g. when
    /// tracking time values stated in nanosecond units, where the minimal accuracy required is a
    /// microsecond, the proper value for `low` would be 1000.
    ///
    /// `low` is the lowest value that can be discerned (distinguished from 0) by the histogram,
    /// and must be a positive integer that is >= 1. It may be internally rounded down to nearest
    /// power of 2. `high` is the highest value to be tracked by the histogram, and must be a
    /// positive integer that is `>= (2 * low)`. `sigfig` Specifies the number of significant
    /// figures to maintain. This is the number of significant decimal digits to which the
    /// histogram will maintain value resolution and separation. Must be a non-negative integer
    /// between 0 and 5.
    pub fn new_with_bounds(low: i64, high: i64, sigfig: u32) -> Result<Histogram<T>, &'static str> {
        // Verify argument validity
        if low < 1 {
            return Err("lowest discernible value must be >= 1");
        }
        if high < 2 * low {
            return Err("highest trackable value must be >= 2 * lowest discernible value");
        }
        if sigfig > 5 {
            return Err("number of significant digits must be between 0 and 5");
        }

        // Given a 3 decimal point accuracy, the expectation is obviously for "+/- 1 unit at 1000".
        // It also means that it's "ok to be +/- 2 units at 2000". The "tricky" thing is that it is
        // NOT ok to be +/- 2 units at 1999. Only starting at 2000. So internally, we need to
        // maintain single unit resolution to 2x 10^decimalPoints.

        // largest value with single unit resolution
        let largest = 2 * 10i64.pow(sigfig);

        let unitMagnitude = ((low as f64).log2() / 2f64.log2()).floor() as usize;
        let unitMagnitudeMask = (1 << unitMagnitude) - 1;

        // We need to maintain power-of-two subBucketCount (for clean direct indexing) that is
        // large enough to provide unit resolution to at least
        // largestValueWithSingleUnitResolution. So figure out
        // largestValueWithSingleUnitResolution's nearest power-of-two (rounded up), and use that:
        let subBucketCountMagnitude = ((largest as f64).log2() / 2f64.log2()).ceil() as usize;
        let subBucketHalfCountMagnitude = if subBucketCountMagnitude > 1 {
            subBucketCountMagnitude
        } else {
            1
        } - 1;
        let subBucketCount = 2usize.pow(subBucketHalfCountMagnitude as u32 + 1);
        let subBucketHalfCount = subBucketCount / 2;
        let subBucketMask = (subBucketCount as i64 - 1) << unitMagnitude;

        let mut h = Histogram {
            autoResize: false,

            highestTrackableValue: high,
            lowestDiscernibleValue: low,
            significantValueDigits: sigfig,

            bucketCount: 0, // set by establishSize below
            subBucketCount: subBucketCount,

            leadingZeroCountBase: 0, // set below, needs establishSize
            subBucketHalfCountMagnitude: subBucketHalfCountMagnitude,

            unitMagnitude: unitMagnitude,
            subBucketHalfCount: subBucketHalfCount,

            subBucketMask: subBucketMask,

            unitMagnitudeMask: unitMagnitudeMask,
            maxValue: 0,
            minNonZeroValue: i64::max_value(),

            // TODO: atomics for maxValue/minNonZeroValue ?
            totalCount: 0,
            counts: Vec::new(), // set by alloc() below
        };

        // determine exponent range needed to support the trackable value with no overflow:
        let len = try!(h.cover(high));

        // Establish leadingZeroCountBase, used in bucketIndexOf() fast path:
        // subtract the bits that would be used by the largest value in bucket 0.
        h.leadingZeroCountBase = (64 - h.unitMagnitude - h.subBucketHalfCountMagnitude - 1) as u32;

        h.alloc(len);
        Ok(h)
    }

    /// Construct a `Histogram` with the same range settings as a given source histogram,
    /// duplicating the source's start/end timestamps (but NOT its contents).
    pub fn new_from<F: num::Num + num::ToPrimitive + Copy>(source: &Histogram<F>) -> Histogram<T> {
        let mut h = Self::new_with_bounds(source.lowestDiscernibleValue,
                                          source.highestTrackableValue,
                                          source.significantValueDigits)
            .unwrap();

        // h.startTime = source.startTime;
        // h.endTime = source.endTime;
        h.autoResize = source.autoResize;
        h.alloc(source.len());
        h
    }

    /// Allocate a counts array of the given size.
    fn alloc(&mut self, len: usize) {
        use std::iter;
        self.counts = iter::repeat(T::zero()).take(len).collect();
    }

    // ********************************************************************************************
    // Recording samples.
    // ********************************************************************************************

    /// Record `value` in the histogram.
    ///
    /// Returns an error if `value` exceeds the highest trackable value and auto-resize is
    /// disabled.
    pub fn record(&mut self, value: i64) -> Result<(), ()> {
        self.record_n(value, T::one())
    }

    /// Record multiple samples for a value in the histogram, adding to the value's current count.
    ///
    /// `count` is the number of occurrences of this value to record. Returns an error if `value`
    /// exceeds the highest trackable value and auto-resize is disabled.
    pub fn record_n(&mut self, value: i64, count: T) -> Result<(), ()> {
        self.alter_n(value, count, true)
    }

    fn alter_n(&mut self, value: i64, count: T, add: bool) -> Result<(), ()> {
        let success = if let Some(c) = self.mut_at(value) {
            if add {
                *c = *c + count;
            } else {
                *c = *c - count;
            }
            true
        } else {
            false
        };

        if !success {
            if !self.autoResize {
                return Err(());
            }

            self.resize(value);

            {
                let c = self.mut_at(value).expect("value should fit after resize");
                if add {
                    *c = *c + count;
                } else {
                    *c = *c - count;
                }
            }

            self.highestTrackableValue = self.highest_equivalent(self.value_for(self.last()));
        }

        self.update_min_max(value);
        if add {
            self.totalCount += count.to_i64().unwrap();
        } else {
            self.totalCount -= count.to_i64().unwrap();
        }
        Ok(())
    }

    /// Record a value in the histogram while correcting for coordinated omission.
    ///
    /// See `record_correct` for further documentation.
    pub fn record_correct(&mut self, value: i64, interval: i64) -> Result<(), ()> {
        self.record_n_correct(value, T::one(), interval)
    }

    /// Record multiple values in the histogram while correcting for coordinated omission.
    ///
    /// To compensate for the loss of sampled values when a recorded value is larger than the
    /// expected interval between value samples, this method will auto-generate and record an
    /// additional series of decreasingly-smaller (down to `interval`) value records.
    ///
    /// Note: This is a at-recording correction method, as opposed to the post-recording correction
    /// method provided by `correct_clone`. The two methods are mutually exclusive, and only one of
    /// the two should be be used on a given data set to correct for the same coordinated omission
    /// issue.
    ///
    /// Returns an error if `value` exceeds the highest trackable value and auto-resize is
    /// disabled.
    pub fn record_n_correct(&mut self, value: i64, count: T, interval: i64) -> Result<(), ()> {
        try!(self.record_n(value, count));
        if interval <= 0 {
            return Ok(());
        }

        let mut missingValue = value - interval;
        while missingValue >= interval {
            try!(self.record_n(missingValue, count));
            missingValue -= interval;
        }
        Ok(())
    }

    // ********************************************************************************************
    // Iterators
    // ********************************************************************************************

    /// Iterate through histogram values by percentile levels.
    ///
    /// The iteration mechanic for this iterator may appear somewhat confusing, but it yields
    /// fairly pleasing output. The iterator starts with a *percentile step size* of
    /// `100/halving_period`. For every iteration, it yields a value whose percentile is that much
    /// greater than the previously emitted percentile (i.e., initially 0, 10, 20, etc.). Once
    /// `halving_period` values have been emitted, the percentile step size is halved, and the
    /// iteration continues.
    ///
    /// The iterator yields a four-value tuple with the following values (in order): the value at
    /// the current iterator step, the percentile of samples at or below that value, the number of
    /// samples recorded at this value, and the number of samples present between the last iterator
    /// value and the current one.
    ///
    /// ```
    /// use hdrsample::Histogram;
    /// let mut hist = Histogram::<u64>::new_with_max(10000, 4).unwrap();
    /// for i in 0..10000 {
    ///     hist += i;
    /// }
    ///
    /// let mut perc = hist.iter_percentiles(1);
    ///
    /// println!("{:?}", hist.iter_percentiles(1).collect::<Vec<_>>());
    ///
    /// assert_eq!(perc.next(), Some((hist.value_at_percentile(0.01), 0.01, 1, 1)));
    /// // step size = 50
    /// assert_eq!(perc.next(), Some((hist.value_at_percentile(50.0), 50.0, 1, 5000 - 1)));
    /// // step size = 25
    /// assert_eq!(perc.next(), Some((hist.value_at_percentile(75.0), 75.0, 1, 2500)));
    /// // step size = 12.5
    /// assert_eq!(perc.next(), Some((hist.value_at_percentile(87.5), 87.5, 1, 1250)));
    /// // step size = 6.25
    /// assert_eq!(perc.next(), Some((hist.value_at_percentile(93.75), 93.75, 1, 625)));
    /// // step size = 3.125
    /// assert_eq!(perc.next(), Some((hist.value_at_percentile(96.88), 96.88, 1, 313)));
    /// // etc...
    /// ```
    pub fn iter_percentiles<'a>
        (&'a self,
         percentileTicksPerHalfDistance: isize)
         -> iterators::HistogramIterator<'a, T, iterators::percentile::Iter<'a, T>> {
        iterators::percentile::Iter::new(self, percentileTicksPerHalfDistance)
    }

    /// Iterates through histogram values using linear value steps. The iteration is performed in
    /// steps of size `step`, each one yielding the count for all values in the preceeding value
    /// range of size `step`. The iterator terminates when all recorded histogram values are
    /// exhausted.
    ///
    /// The iterator yields a four-value tuple with the following values (in order): the value at
    /// the current iterator step, the percentile of samples at or below that value, the number of
    /// samples recorded at this value, and the number of samples present between the last iterator
    /// value and the current one.
    ///
    /// ```
    /// use hdrsample::Histogram;
    /// let mut hist = Histogram::<u64>::new_with_max(1000, 3).unwrap();
    /// hist += 100;
    /// hist += 500;
    /// hist += 800;
    /// hist += 850;
    ///
    /// let mut perc = hist.iter_linear(100);
    /// assert_eq!(perc.next(), Some((99, hist.percentile_below(99), 0, 0)));
    /// assert_eq!(perc.next(), Some((199, hist.percentile_below(199), 0, 1)));
    /// assert_eq!(perc.next(), Some((299, hist.percentile_below(299), 0, 0)));
    /// assert_eq!(perc.next(), Some((399, hist.percentile_below(399), 0, 0)));
    /// assert_eq!(perc.next(), Some((499, hist.percentile_below(499), 0, 0)));
    /// assert_eq!(perc.next(), Some((599, hist.percentile_below(599), 0, 1)));
    /// assert_eq!(perc.next(), Some((699, hist.percentile_below(699), 0, 0)));
    /// assert_eq!(perc.next(), Some((799, hist.percentile_below(799), 0, 0)));
    /// assert_eq!(perc.next(), Some((899, hist.percentile_below(899), 0, 2)));
    /// assert_eq!(perc.next(), None);
    /// ```
    pub fn iter_linear<'a>
        (&'a self,
         step: i64)
         -> iterators::HistogramIterator<'a, T, iterators::linear::Iter<'a, T>> {
        iterators::linear::Iter::new(self, step)
    }

    /// Iterates through histogram values at logarithmically increasing levels. The iteration is
    /// performed in steps that start at `start` and increase exponentially according to `exp`. The
    /// iterator terminates when all recorded histogram values are exhausted.
    ///
    /// The iterator yields a four-value tuple with the following values (in order): the value at
    /// the current iterator step, the percentile of samples at or below that value, the number of
    /// samples recorded at this value, and the number of samples present between the last iterator
    /// value and the current one.
    ///
    /// ```
    /// use hdrsample::Histogram;
    /// let mut hist = Histogram::<u64>::new_with_max(1000, 3).unwrap();
    /// hist += 100;
    /// hist += 500;
    /// hist += 800;
    /// hist += 850;
    ///
    /// let mut perc = hist.iter_log(1, 10.0);
    /// assert_eq!(perc.next(), Some((0, hist.percentile_below(0), 0, 0)));
    /// assert_eq!(perc.next(), Some((9, hist.percentile_below(9), 0, 0)));
    /// assert_eq!(perc.next(), Some((99, hist.percentile_below(99), 0, 0)));
    /// assert_eq!(perc.next(), Some((999, hist.percentile_below(999), 0, 4)));
    /// assert_eq!(perc.next(), None);
    /// ```
    pub fn iter_log<'a>(&'a self,
                        start: i64,
                        exp: f64)
                        -> iterators::HistogramIterator<'a, T, iterators::log::Iter<'a, T>> {
        iterators::log::Iter::new(self, start, exp)
    }

    /// Iterates through all recorded histogram values using the finest granularity steps supported
    /// by the underlying representation. The iteration steps through all non-zero recorded value
    /// counts, and terminates when all recorded histogram values are exhausted.
    ///
    /// The iterator yields a four-value tuple with the following values (in order): the value at
    /// the current iterator step, the percentile of samples at or below that value, the number of
    /// samples recorded at this value, and the number of samples present between the last iterator
    /// value and the current one.
    ///
    /// ```
    /// use hdrsample::Histogram;
    /// let mut hist = Histogram::<u64>::new_with_max(1000, 3).unwrap();
    /// hist += 100;
    /// hist += 500;
    /// hist += 800;
    /// hist += 850;
    ///
    /// let mut perc = hist.iter_recorded();
    /// assert_eq!(perc.next(), Some((100, hist.percentile_below(100), 1, 1)));
    /// assert_eq!(perc.next(), Some((500, hist.percentile_below(500), 1, 1)));
    /// assert_eq!(perc.next(), Some((800, hist.percentile_below(800), 1, 1)));
    /// assert_eq!(perc.next(), Some((850, hist.percentile_below(850), 1, 1)));
    /// assert_eq!(perc.next(), None);
    /// ```
    pub fn iter_recorded<'a>
        (&'a self)
         -> iterators::HistogramIterator<'a, T, iterators::recorded::Iter<'a, T>> {
        iterators::recorded::Iter::new(self)
    }

    /// Iterates through all histogram values using the finest granularity steps supported by the
    /// underlying representation. The iteration steps through all possible unit value levels,
    /// regardless of whether or not there were recorded values for that value level, and
    /// terminates when all recorded histogram values are exhausted.
    ///
    /// The iterator yields a four-value tuple with the following values (in order): the value at
    /// the current iterator step, the percentile of samples at or below that value, the number of
    /// samples recorded at this value, and the number of samples present between the last iterator
    /// value and the current one.
    ///
    /// ```
    /// use hdrsample::Histogram;
    /// let mut hist = Histogram::<u64>::new_with_max(10, 1).unwrap();
    /// hist += 1;
    /// hist += 5;
    /// hist += 8;
    ///
    /// let mut perc = hist.iter_all();
    /// assert_eq!(perc.next(), Some((0, 0.0, 0, 0)));
    /// assert_eq!(perc.next(), Some((1, hist.percentile_below(1), 1, 1)));
    /// assert_eq!(perc.next(), Some((2, hist.percentile_below(2), 0, 0)));
    /// assert_eq!(perc.next(), Some((3, hist.percentile_below(3), 0, 0)));
    /// assert_eq!(perc.next(), Some((4, hist.percentile_below(4), 0, 0)));
    /// assert_eq!(perc.next(), Some((5, hist.percentile_below(5), 1, 1)));
    /// assert_eq!(perc.next(), Some((6, hist.percentile_below(6), 0, 0)));
    /// assert_eq!(perc.next(), Some((7, hist.percentile_below(7), 0, 0)));
    /// assert_eq!(perc.next(), Some((8, hist.percentile_below(8), 1, 1)));
    /// assert_eq!(perc.next(), Some((9, hist.percentile_below(9), 0, 0)));
    /// assert_eq!(perc.next(), Some((10, 100.0, 0, 0)));
    /// ```
    pub fn iter_all<'a>(&'a self) -> iterators::HistogramIterator<'a, T, iterators::all::Iter> {
        iterators::all::Iter::new(self)
    }

    // ********************************************************************************************
    // Data statistics
    // ********************************************************************************************

    /// Get the lowest recorded value level in the histogram.
    /// If the histogram has no recorded values, the value returned will be 0.
    pub fn min(&self) -> i64 {
        if self.totalCount == 0 || self[0usize] != T::zero() {
            0
        } else {
            self.min_nz()
        }
    }

    /// Get the highest recorded value level in the histogram.
    /// If the histogram has no recorded values, the value returned is undefined.
    pub fn max(&self) -> i64 {
        if self.maxValue == 0 {
            0
        } else {
            self.highest_equivalent(self.maxValue)
        }
    }

    /// Get the lowest recorded non-zero value level in the histogram.
    /// If the histogram has no recorded values, the value returned is `i64::max_value()`.
    pub fn min_nz(&self) -> i64 {
        if self.minNonZeroValue == i64::max_value() {
            i64::max_value()
        } else {
            self.lowest_equivalent(self.minNonZeroValue)
        }
    }

    /// Determine if two values are equivalent with the histogram's resolution. Equivalent here
    /// means that value samples recorded for any two equivalent values are counted in a common
    /// total count.
    pub fn equivalent(&self, value1: i64, value2: i64) -> bool {
        self.lowest_equivalent(value1) == self.lowest_equivalent(value2)
    }

    /// Get the computed mean value of all recorded values in the histogram.
    pub fn mean(&self) -> f64 {
        if self.totalCount == 0 {
            return 0.0;
        }

        self.iter_recorded().fold(0.0f64, |total, (v, _, c, _)| {
            total +
            self.median_equivalent(v) as f64 * c.to_i64().unwrap() as f64 / self.totalCount as f64
        })
    }

    /// Get the computed standard deviation of all recorded values in the histogram
    pub fn stdev(&self) -> f64 {
        if self.totalCount == 0 {
            return 0.0;
        }

        let mean = self.mean();
        let geom_dev_tot = self.iter_recorded().fold(0.0f64, |gdt, (v, _, _, sc)| {
            let dev = self.median_equivalent(v) as f64 - mean;
            gdt + (dev * dev) * sc as f64
        });

        (geom_dev_tot / self.totalCount as f64).sqrt()
    }

    /// Get the value at a given percentile.
    ///
    /// When the given percentile is > 0.0, the value returned is the value that the given
    /// percentage of the overall recorded value entries in the histogram are either smaller than
    /// or equivalent to. When the given percentile is 0.0, the value returned is the value that
    /// all value entries in the histogram are either larger than or equivalent to.
    ///
    /// Two values are considered "equivalent" if `self.equivalent` would return true.
    pub fn value_at_percentile(&self, percentile: f64) -> i64 {
        use std::cmp;

        // Truncate down to 100%
        let percentile = if percentile > 100.0 {
            100.0
        } else {
            percentile
        };

        // round to nearest
        let countAtPercentile = (((percentile / 100.0) * self.totalCount as f64) + 0.5) as i64;

        // Make sure we at least reach the first recorded entry
        let countAtPercentile = cmp::max(countAtPercentile, 1);

        let mut totalToCurrentIndex = 0i64;
        for i in 0..self.len() {
            totalToCurrentIndex += self[i].to_i64().unwrap();
            if totalToCurrentIndex >= countAtPercentile {
                let valueAtIndex = self.value_for(i);
                return if percentile == 0.0 {
                    self.lowest_equivalent(valueAtIndex)
                } else {
                    self.highest_equivalent(valueAtIndex)
                };
            }
        }

        0
    }

    /// Get the percentile of samples at and below a given value.
    ///
    /// The percentile returned is the percentile of values recorded in the histogram that are
    /// smaller than or equivalent to the given value.
    ///
    /// Two values are considered "equivalent" if `self.equivalent` would return true.
    pub fn percentile_below(&self, value: i64) -> f64 {
        use std::cmp;

        if self.totalCount == 0 {
            return 100.0;
        }

        let targetIndex = cmp::min(self.index_for(value), self.last() as isize) as usize;
        let totalToCurrentIndex: i64 =
            (0..(targetIndex + 1)).map(|i| self[i].to_i64().unwrap()).fold(0, |t, v| t + v);
        (100 * totalToCurrentIndex) as f64 / self.totalCount as f64
    }

    /// Get the count of recorded values within a range of value levels (inclusive to within the
    /// histogram's resolution).
    ///
    /// `low` gives the lower value bound on the range for which to provide the recorded count.
    /// Will be rounded down with `lowest_equivalent`. Similarly, `high` gives the higher value
    /// bound on the range, and will be rounded up with `highest_equivalent`. The function returns
    /// the total count of values recorded in the histogram within the value range that is `>=
    /// lowest_equivalent(low)` and `<= highest_equivalent(high)`.
    ///
    /// May fail if the given values are out of bounds.
    pub fn count_between(&self, low: i64, high: i64) -> Result<i64, ()> {
        use std::cmp;
        let lowIndex = cmp::max(0, self.index_for(low)) as usize;
        let highIndex = cmp::min(self.index_for(high), self.last() as isize) as usize;
        Ok((lowIndex..(highIndex + 1)).map(|i| self[i].to_i64().unwrap()).fold(0, |t, v| t + v))
    }

    /// Get the count of recorded values at a specific value (to within the histogram resolution at
    /// the value level).
    ///
    /// The count is cumputed across values recorded in the histogram that are within the value
    /// range that is `>= lowest_equivalent(value)` and `<= highest_equivalent(value)`.
    ///
    /// May fail if the given value is out of bounds.
    pub fn count_at(&self, value: i64) -> Result<T, ()> {
        use std::cmp;
        Ok(self[cmp::min(cmp::max(0, self.index_for(value)), self.last() as isize) as usize])
    }

    // ********************************************************************************************
    // Public helpers
    // ********************************************************************************************

    /// Computes the matching histogram value for the given histogram bin.
    pub fn value_for(&self, index: usize) -> i64 {
        let mut bucketIndex = (index >> self.subBucketHalfCountMagnitude) as isize - 1;
        let mut subBucketIndex =
            ((index & (self.subBucketHalfCount - 1)) + self.subBucketHalfCount) as isize;
        if bucketIndex < 0 {
            subBucketIndex -= self.subBucketHalfCount as isize;
            bucketIndex = 0;
        }
        self.value_from_loc(bucketIndex as usize, subBucketIndex as usize)
    }

    /// Get the lowest value that is equivalent to the given value within the histogram's
    /// resolution. Equivalent here means that value samples recorded for any two equivalent values
    /// are counted in a common total count.
    pub fn lowest_equivalent(&self, value: i64) -> i64 {
        let bucketIndex = self.bucket_for(value);
        let subBucketIndex = self.sub_bucket_for(value, bucketIndex);
        self.value_from_loc(bucketIndex, subBucketIndex)
    }

    /// Get the highest value that is equivalent to the given value within the histogram's
    /// resolution. Equivalent here means that value samples recorded for any two equivalent values
    /// are counted in a common total count.
    ///
    /// Note that the return value is capped at `i64::max_value()`.
    pub fn highest_equivalent(&self, value: i64) -> i64 {
        if value == i64::max_value() {
            i64::max_value()
        } else {
            self.next_non_equivalent(value) - 1
        }
    }

    /// Get a value that lies in the middle (rounded up) of the range of values equivalent the
    /// given value. Equivalent here means that value samples recorded for any two equivalent
    /// values are counted in a common total count.
    ///
    /// Note that the return value is capped at `i64::max_value()`.
    pub fn median_equivalent(&self, value: i64) -> i64 {
        match self.lowest_equivalent(value).overflowing_add(self.equivalent_range(value) >> 1) {
            (_, of) if of => i64::max_value(),
            (v, _) => v,
        }
    }

    /// Get the next value that is *not* equivalent to the given value within the histogram's
    /// resolution. Equivalent means that value samples recorded for any two equivalent values are
    /// counted in a common total count.
    ///
    /// Note that the return value is capped at `i64::max_value()`.
    pub fn next_non_equivalent(&self, value: i64) -> i64 {
        match self.lowest_equivalent(value).overflowing_add(self.equivalent_range(value)) {
            (_, of) if of => i64::max_value(),
            (v, _) => v,
        }
    }

    /// Get the size (in value units) of the range of values that are equivalent to the given value
    /// within the histogram's resolution. Equivalent here means that value samples recorded for
    /// any two equivalent values are counted in a common total count.
    pub fn equivalent_range(&self, value: i64) -> i64 {
        let bucketIndex = self.bucket_for(value);
        let subBucketIndex = self.sub_bucket_for(value, bucketIndex);
        // calculate distance to next value
        1i64 <<
        (self.unitMagnitude +
         if subBucketIndex >= self.subBucketCount {
            bucketIndex + 1
        } else {
            bucketIndex
        })
    }

    // ********************************************************************************************
    // Internal helpers
    // ********************************************************************************************

    /// Compute the lowest (and therefore highest precision) bucket index that can represent the
    /// value.
    #[inline]
    fn bucket_for(&self, value: i64) -> usize {
        // Calculates the number of powers of two by which the value is greater than the biggest
        // value that fits in bucket 0. This is the bucket index since each successive bucket can
        // hold a value 2x greater. The mask maps small values to bucket 0.
        (self.leadingZeroCountBase - (value | self.subBucketMask).leading_zeros()) as usize
    }

    #[inline]
    /// Compute the position inside a bucket at which the given value should be recorded.
    fn sub_bucket_for(&self, value: i64, bucketIndex: usize) -> usize {
        // For bucketIndex 0, this is just value, so it may be anywhere in 0 to subBucketCount. For
        // other bucketIndex, this will always end up in the top half of subBucketCount: assume
        // that for some bucket k > 0, this calculation will yield a value in the bottom half of 0
        // to subBucketCount. Then, because of how buckets overlap, it would have also been in the
        // top half of bucket k-1, and therefore would have returned k-1 in bucketIndexOf(). Since
        // we would then shift it one fewer bits here, it would be twice as big, and therefore in
        // the top half of subBucketCount.
        // TODO: >>> ?
        (value >> (bucketIndex + self.unitMagnitude)) as usize
    }

    #[inline]
    fn value_from_loc(&self, bucketIndex: usize, subBucketIndex: usize) -> i64 {
        (subBucketIndex as i64) << (bucketIndex + self.unitMagnitude)
    }

    /// Find the number of buckets needed such that `value` is representable.
    fn buckets_to_cover(&self, value: i64) -> usize {
        // the k'th bucket can express from 0 * 2^k to subBucketCount * 2^k in units of 2^k
        let mut smallestUntrackableValue = (self.subBucketCount as i64) << self.unitMagnitude;

        // always have at least 1 bucket
        let mut bucketsNeeded = 1;
        while smallestUntrackableValue <= value {
            if smallestUntrackableValue > i64::max_value() / 2 {
                // next shift will overflow, meaning that bucket could represent values up to ones
                // greater than i64::max_value, so it's the last bucket
                return bucketsNeeded + 1;
            }
            smallestUntrackableValue <<= 1;
            bucketsNeeded += 1;
        }
        bucketsNeeded
    }

    /// Compute the actual number of bins to use for the given bucket count (that is, including the
    /// sub-buckets within each top-level bucket).
    ///
    /// If we have `N` such that `subBucketCount * 2^N > high`, we need storage for `N+1` buckets,
    /// each with enough slots to hold the top half of the `subBucketCount` (the lower half is
    /// covered by previous buckets), and the +1 being used for the lower half of the 0'th bucket.
    /// Or, equivalently, we need 1 more bucket to capture the max value if we consider the
    /// sub-bucket length to be halved.
    fn num_bins(&self, numberOfBuckets: usize) -> usize {
        (numberOfBuckets + 1) * (self.subBucketCount / 2)
    }

    /// Compute the number of buckets needed to cover the given value, as well as the total number
    /// of bins needed to cover that range.
    ///
    /// May fail if `high` is not at least twice the lowest discernible value.
    fn cover(&mut self, high: i64) -> Result<usize, &'static str> {
        if high < 2 * self.lowestDiscernibleValue {
            return Err("highest trackable value cannot be < (2 * lowest discernible value)");
        }

        // establish counts array length:
        let bucketsNeeded = self.buckets_to_cover(high);
        let countsArrayLength = self.num_bins(bucketsNeeded);

        // establish exponent range needed to support the trackable value with no overflow:
        self.bucketCount = bucketsNeeded;

        // establish the new highest trackable value:
        self.highestTrackableValue = high;

        Ok(countsArrayLength)
    }

    /// Resize the underlying counts array such that it can cover the given `high` value.
    fn resize(&mut self, high: i64) {
        // figure out how large the sample tracker now needs to be
        let len = self.cover(high).unwrap();

        // expand counts to also hold the new counts
        self.counts.resize(len, T::zero());
    }

    /// Set internally tracked maxValue to new value if new value is greater than current one.
    fn update_max(&mut self, value: i64) {
        let internalValue = value | self.unitMagnitudeMask; // Max unit-equivalent value
        if internalValue > self.maxValue {
            self.maxValue = internalValue;
        }
    }

    /// Set internally tracked minNonZeroValue to new value if new value is smaller than current
    /// one.
    fn update_min(&mut self, value: i64) {
        if value <= self.unitMagnitudeMask {
            return; // Unit-equivalent to 0.
        }

        let internalValue = value & !self.unitMagnitudeMask; // Min unit-equivalent value
        if internalValue < self.minNonZeroValue {
            self.minNonZeroValue = internalValue;
        }
    }

    fn update_min_max(&mut self, value: i64) {
        if value > self.maxValue {
            self.update_max(value);
        }
        if value < self.minNonZeroValue && value != 0 {
            self.update_min(value);
        }
    }

    fn reset_max(&mut self, max: i64) {
        self.maxValue = max | self.unitMagnitudeMask; // Max unit-equivalent value
    }

    fn reset_min(&mut self, min: i64) {
        let internalValue = min & !self.unitMagnitudeMask; // Min unit-equivalent value
        self.minNonZeroValue = if min == i64::max_value() {
            min
        } else {
            internalValue
        };
    }

    fn restat(&mut self, until: usize) {
        self.reset_max(0);
        self.reset_min(i64::max_value());

        let mut max_i = None;
        let mut min_i = None;
        let mut totalCount = 0;
        for i in 0..until {
            let count = self[i];
            if count != T::zero() {
                totalCount += count.to_i64().unwrap();
                max_i = Some(i);
                if min_i.is_none() && i != 0 {
                    min_i = Some(i);
                }
            }
        }

        if let Some(max_i) = max_i {
            let max = self.highest_equivalent(self.value_for(max_i));
            self.update_max(max);
        }
        if let Some(min_i) = min_i {
            let min = self.value_for(min_i);
            self.update_min(min);
        }

        self.totalCount = totalCount;
    }
}

// ********************************************************************************************
// Trait implementations
// ********************************************************************************************

impl<T: num::Num + num::ToPrimitive + Copy> Index<usize> for Histogram<T> {
    type Output = T;
    fn index(&self, index: usize) -> &Self::Output {
        &self.counts[index]
    }
}

impl<T: num::Num + num::ToPrimitive + Copy> IndexMut<usize> for Histogram<T> {
    fn index_mut(&mut self, index: usize) -> &mut Self::Output {
        &mut self.counts[index]
    }
}

impl<T: num::Num + num::ToPrimitive + Copy> Clone for Histogram<T> {
    fn clone(&self) -> Self {
        let mut h = Histogram::new_from(self);
        h += self;
        h
    }
}

// make it more ergonomic to add and subtract histograms
impl<'a, T: num::Num + num::ToPrimitive + Copy> AddAssign<&'a Histogram<T>> for Histogram<T> {
    fn add_assign(&mut self, source: &'a Histogram<T>) {
        self.add(source).unwrap();
    }
}

impl<'a, T: num::Num + num::ToPrimitive + Copy> SubAssign<&'a Histogram<T>> for Histogram<T> {
    fn sub_assign(&mut self, other: &'a Histogram<T>) {
        self.subtract(other).unwrap();
    }
}

// make it more ergonomic to record samples
impl<T: num::Num + num::ToPrimitive + Copy> AddAssign<i64> for Histogram<T> {
    fn add_assign(&mut self, value: i64) {
        self.record(value).unwrap();
    }
}

// allow comparing histograms
impl<T: num::Num + num::ToPrimitive + Copy, F: num::Num + num::ToPrimitive + Copy> PartialEq<Histogram<F>> for Histogram<T> {
    fn eq(&self, other: &Histogram<F>) -> bool {
        if self.lowestDiscernibleValue != other.lowestDiscernibleValue ||
           self.significantValueDigits != other.significantValueDigits {
            return false;
        }
        if self.totalCount != other.totalCount {
            return false;
        }
        if self.max() != other.max() {
            return false;
        }
        if self.min_nz() != other.min_nz() {
            return false;
        }
        (0..self.len()).all(|i| self[i].to_i64() == other[i].to_i64())
    }
}

// /**
//  * Indicate whether or not the histogram is capable of supporting auto-resize functionality.
//  * Note that this is an indication that enabling auto-resize by calling setAutoResize() is
//  * allowed, and NOT that the histogram will actually auto-resize. Use isAutoResize() to
//  * determine if the histogram is in auto-resize mode.
//  * @return autoResize setting
//  */
// public boolean supportsAutoResize() { return true; }

// TODO: shift
// TODO: hash
// TODO: serialization
// TODO: encoding/decoding
// TODO: timestamps and tags
// TODO: textual output
